#include "Conversions.h"

//---------------------------------------------------------------------------
//  
//---------------------------------------------------------------------------

void QuaternionToMatrix(const CQuaternion &Quaternion,CMatrix44 &Matrix)
	{
	/*
	If Quaternion is guaranteed to be a unit CQuaternion, s will always
	be 1.  In that case, this calculation can be optimized out.
	*/
	float norm,s,xx,yy,zz,xy,xz,yz,wx,wy,wz;
	norm = Quaternion.X*Quaternion.X + Quaternion.Y*Quaternion.Y + Quaternion.Z*Quaternion.Z + Quaternion.W*Quaternion.W;
	s = (norm > 0) ? 2/norm : 0;
	
	/*
	Precalculate coordinate products
	*/
	xx = Quaternion.X * Quaternion.X * s;
	yy = Quaternion.Y * Quaternion.Y * s;
	zz = Quaternion.Z * Quaternion.Z * s;
	xy = Quaternion.X * Quaternion.Y * s;
	xz = Quaternion.X * Quaternion.Z * s;
	yz = Quaternion.Y * Quaternion.Z * s;
	wx = Quaternion.W * Quaternion.X * s;
	wy = Quaternion.W * Quaternion.Y * s;
	wz = Quaternion.W * Quaternion.Z * s;

	/*
	Calculate 3x3 matrix from orthonormal basis
	*/

	/*
	x axis
	*/
	Matrix[kX][kX] = 1.0f - (yy + zz);
	Matrix[kY][kX] = xy + wz;
	Matrix[kZ][kX] = xz - wy;
	
	/*
	y axis
	*/
	Matrix[kX][kY] = xy - wz;
	Matrix[kY][kY] = 1.0f - (xx + zz);
	Matrix[kZ][kY] = yz + wx;
	
	/*
	z axis
	*/
	Matrix[kX][kZ] = xz + wy;
	Matrix[kY][kZ] = yz - wx;
	Matrix[kZ][kZ] = 1.0f - (xx + yy);

	/*
	4th row and column of 4x4 matrix
	Translation and scale are not stored in quaternions, so these
	values are set to default (no scale, no translation).
	For systems where Matrix comes pre-loaded with scale and translation
	factors, this code can be excluded.
	*/
	Matrix[kW][kX] = Matrix[kW][kY] = Matrix[kW][kZ] = Matrix[kX][kW] = Matrix[kY][kW] = Matrix[kZ][kW] = 0.0;
	Matrix[kW][kW] = 1.0f;
	}

//---------------------------------------------------------------------------
//  
//---------------------------------------------------------------------------

void MatrixToQuaternion(const CMatrix44 &Matrix,CQuaternion &Quaternion)
	{
	/*
	This code can be optimized for Matrix[kW][kW] = 1, which 
	should always be true.  This optimization is excluded
	here for clarity.
	*/
	
	float Tr = Matrix[kX][kX] + Matrix[kY][kY] + Matrix[kZ][kZ] + Matrix[kW][kW],fourD;
	int i,j,k;
	
	/*
	w >= 0.5 ?
	*/
	if (Tr >= 1.0f)
		{
		fourD = 2.0f*(float)sqrt(Tr);
		Quaternion.W = fourD/4.0f;
		Quaternion.X = (Matrix[kZ][kY] - Matrix[kY][kZ])/fourD;
		Quaternion.Y = (Matrix[kX][kZ] - Matrix[kZ][kX])/fourD;
		Quaternion.Z = (Matrix[kY][kX] - Matrix[kX][kY])/fourD;
		}
	else
		{
		/*
		Find the largest component.  
		*/
		if (Matrix[kX][kX] > Matrix[kY][kY])
			{
			i = kX;
			}
		else
			{
			i = kY;
			}
		if (Matrix[kZ][kZ] > Matrix[i][i])
			{
			i = kZ;
			}
		
		/*
		Set j and k to point to the next two components
		*/
		j = (i+1)%3;
		k = (j+1)%3;

		/*
		fourD = 4 * largest component
		*/
		fourD = 2.0f*(float)sqrt(Matrix[i][i] - Matrix[j][j] - Matrix[k][k] + 1.0 );

		/*
		Set the largest component
		*/
		switch (i)
			{
			case 0:Quaternion.X=fourD/4.0f;
			case 1:Quaternion.Y=fourD/4.0f;
			case 2:Quaternion.Z=fourD/4.0f;
			case 3:Quaternion.W=fourD/4.0f;
			}
		
		/*
		Calculate remaining components
		*/
		switch (j)
			{
			case 0:Quaternion.X=(Matrix[j][i] + Matrix[i][j])/fourD;
			case 1:Quaternion.Y=(Matrix[j][i] + Matrix[i][j])/fourD;
			case 2:Quaternion.Z=(Matrix[j][i] + Matrix[i][j])/fourD;
			case 3:Quaternion.W=(Matrix[j][i] + Matrix[i][j])/fourD;
			}
		switch (k)
			{
			case 0:Quaternion.X=(Matrix[j][i] + Matrix[i][k])/fourD;
			case 1:Quaternion.Y=(Matrix[j][i] + Matrix[i][k])/fourD;
			case 2:Quaternion.Z=(Matrix[j][i] + Matrix[i][k])/fourD;
			case 3:Quaternion.W=(Matrix[j][i] + Matrix[i][k])/fourD;
			}
		Quaternion.W = (Matrix[k][j] - Matrix[j][k])/fourD;
		}
	}

