#pragma once
#ifdef USE_TEMPLATE_METAPROGRAM 
#pragma message("Using template meta-programming....")
#include <algorithm> // std::swap


//---------------------------------------------------------------------------
//  Matriz identidade
//---------------------------------------------------------------------------

template <class Mtx,unsigned Size> struct IdMtx
	{
	static inline void eval( Mtx& mtx )
		{
		IdMtxImpl< Mtx, Size, 0, 0, 0 >::eval( mtx );
		}
	};

// Assigns each element of the matrix
template <class Mtx,unsigned Size,unsigned Column,unsigned Row,unsigned I> struct IdMtxImpl
	{
	enum
		{
		NxtI = I + 1,          // Counter
		NxtR = NxtI % Size,       // Row (inner loop)
		NxtC = NxtI / Size % Size    // Column (outer loop)
		};
	static inline void eval( Mtx& mtx )
		{
		mtx[ Column ][ Row ] = ( Column == Row ) ? 1.0 : 0.0;
		IdMtxImpl< Mtx, Size, NxtC, NxtR, NxtI >::eval( mtx );
		}
	};

// Specialize for 3x3 and 4x4 matrix 
template<> struct IdMtxImpl< CMatrix33, 3, 0, 0, 3*3 >
	{
	static inline void eval( CMatrix33& ) {}
	};
template<> struct IdMtxImpl< CMatrix44, 4, 0, 0, 4*4 >
	{
	static inline void eval( CMatrix44& ) {}
	};

// Make the template appear like a function
#define IdentityMtxT( MtxType, Mtx, Size ) IdMtx< MtxType, Size >::eval( Mtx )

//---------------------------------------------------------------------------
//  Initialize matrix
//---------------------------------------------------------------------------

// Size is matrix size
template< class Mtx, unsigned Size, class F, int Init > struct InitMtx
	{
	static inline void eval( Mtx& mtx )
		{
		InitMtxImpl< Mtx, Size, F, Init, 0, 0, 0 >::eval( mtx );
		}
	};

// Assigns each element of the matrix
template <class Mtx, unsigned Size, class F, int Init, unsigned Column, unsigned Row, unsigned I > struct InitMtxImpl
	{
	enum
		{
		NxtI = I + 1,          // Counter
		NxtR = NxtI % Size,       // Row (inner loop)
		NxtC = NxtI / Size % Size    // Column (outer loop)
		};
	static inline void eval( Mtx& mtx )
		{
		mtx[ Column ][ Row ] = static_cast< F >( Init );
		InitMtxImpl< Mtx, Size, F, Init, NxtC, NxtR, NxtI >::eval( mtx );
		}
	};

// Specialize for 3x3 and 4x4 matrix for initial values of zero and one
template<> struct InitMtxImpl< CMatrix33, 3, float, 0, 0, 0, 3*3 >
	{
	static inline void eval( CMatrix33& ) {}
	};
template<> struct InitMtxImpl< CMatrix33, 3, float, 1, 0, 0, 3*3 >
	{
	static inline void eval( CMatrix33& ) {}
	};
template<> struct InitMtxImpl< CMatrix44, 4, float, 0, 0, 0, 4*4 >
	{
	static inline void eval( CMatrix44& ) {}
	};
template<> struct InitMtxImpl< CMatrix44, 4, float, 1, 0, 0, 4*4 >
	{
	static inline void eval( CMatrix44& ) {}
	};

// Make the template appear like a function
#define InitMtxT( MtxType, Mtx, Size, F, Init ) InitMtx< MtxType, Size, F, Init >::eval( Mtx )

//---------------------------------------------------------------------------
//  Transpose matrix
//---------------------------------------------------------------------------

// Templatized transpose; Size is matrix size
template< class Mtx, unsigned Size > struct TransMtx
	{
	static inline void eval( Mtx& mtx )
		{
		TransMtxImpl< Mtx, Size, 0, 1, 0 >::eval( mtx );
		}
	};

template< class Mtx, unsigned Size, unsigned Column, unsigned Row, unsigned I > struct TransMtxImpl
	{
	enum
		{
		NxtI = I + 1,
		NxtC = NxtI / Size % Size,
		NxtR = ( NxtI % Size ) + NxtC + 1
		};
	static inline void eval( Mtx& mtx )
		{
		if( Row < Size )
		std::swap( mtx[ Column ][ Row ], mtx[ Row ][ Column ] );
		TransMtxImpl< Mtx, Size, NxtC, NxtR, NxtI >::eval( mtx );
		}
	};

// Specialize for 3x3 and 4x4 matrix
template<> struct TransMtxImpl< CMatrix33, 3, 0, 1, 3*3 >
	{
	static inline void eval( CMatrix33& ) {}
	};
template<> struct TransMtxImpl< CMatrix44, 4, 0, 1, 4*4 >
	{
	static inline void eval( CMatrix44& ) {}
	};

// Make the template appear like a function 
#define TransMtxT( MtxType, Mtx, Size ) TransMtx< MtxType, Size >::eval( Mtx )

//---------------------------------------------------------------------------
//  Multiplicao de matrizes
//---------------------------------------------------------------------------

template< class Mtx, unsigned Size > struct MultMtx
	{
	static inline void eval( Mtx& r, const Mtx& a, const Mtx& b )
		{
		InitMtxT( Mtx, r, Size, float, 0 ); // clear the result
		MultMtxImpl< Mtx, Size, 0, 0, 0, 0 >::eval( r, a, b );
		}
	};

template< class Mtx, unsigned Size, unsigned Column, unsigned Row, unsigned K, unsigned I > struct MultMtxImpl
	{
	enum
		{
		NxtI = I + 1,           // Counter
		NxtK = NxtI % Size,        // Internal loop
		NxtC = NxtI / Size % Size,    // Column
		NxtR = NxtI / Size / Size % Size // Row
		};
	static inline void eval( Mtx& r, const Mtx& a, const Mtx& b )
		{
		r[ Column ][ Row ] += a[ K ][ Row ] * b[ Column ][ K ];
		MultMtxImpl< Mtx, Size, NxtC, NxtR, NxtK, NxtI >::eval( r,a,b );
		}
	};

// Specialize for 3x3 and 4x4 matrix
template<> struct MultMtxImpl< CMatrix33, 3, 0, 0, 0, 3*3*3 >
	{
	static inline void eval( CMatrix33&, const CMatrix33&, const CMatrix33& ) {}
	};
template<> struct MultMtxImpl< CMatrix44, 4, 0, 0, 0, 4*4*4 >
	{
	static inline void eval( CMatrix44&, const CMatrix44&, const CMatrix44& ) {}
	};

// Make the template appear like a function 
#define MultMtxT( MtxType, r, a, b, Size ) MultMtx< MtxType, Size >::eval( r, a, b )

#endif